# How to manually call the SIGINT handler on Windows without using raise() or
# similar.

if os.name == "nt":
    import cffi
    ffi = cffi.FFI()
    ffi.cdef("""
        void* WINAPI GetProcAddress(void* hModule, char* lpProcName);
        typedef void (*PyOS_sighandler_t)(int);
    """)
    kernel32 = ffi.dlopen("kernel32.dll")
    PyOS_getsig_ptr = kernel32.GetProcAddress(
        ffi.cast("void*", sys.dllhandle), b"PyOS_getsig")
    PyOS_getsig = ffi.cast("PyOS_sighandler_t (*)(int)", PyOS_getsig_ptr)


    import signal
    PyOS_getsig(signal.SIGINT)(signal.SIGINT)



from contextlib import contextmanager
import ssl

client_ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
client_ctx.check_hostname = False
client_ctx.verify_mode = ssl.CERT_NONE

cinb = ssl.MemoryBIO()
coutb = ssl.MemoryBIO()
cso = client_ctx.wrap_bio(cinb, coutb)

server_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
server_ctx.load_cert_chain("server.crt", "server.key", "xxxx")
sinb = ssl.MemoryBIO()
soutb = ssl.MemoryBIO()
sso = server_ctx.wrap_bio(sinb, soutb, server_side=True)

@contextmanager
def expect(etype):
    try:
        yield
    except etype:
        pass
    else:
        raise AssertionError("expected {}".format(etype))

with expect(ssl.SSLWantReadError):
    cso.do_handshake()
assert not cinb.pending
assert coutb.pending

with expect(ssl.SSLWantReadError):
    sso.do_handshake()
assert not sinb.pending
assert not soutb.pending

# A trickle is not enough
# sinb.write(coutb.read(1))
# with expect(ssl.SSLWantReadError):
#     cso.do_handshake()
# with expect(ssl.SSLWantReadError):
#     sso.do_handshake()

sinb.write(coutb.read())
# Now it should be able to respond
with expect(ssl.SSLWantReadError):
    sso.do_handshake()
assert soutb.pending

cinb.write(soutb.read())
with expect(ssl.SSLWantReadError):
    cso.do_handshake()

sinb.write(coutb.read())
# server done!
sso.do_handshake()
assert soutb.pending

# client done!
cinb.write(soutb.read())
cso.do_handshake()

cso.write(b"hello")
sinb.write(coutb.read())
assert sso.read(10) == b"hello"
with expect(ssl.SSLWantReadError):
    sso.read(10)

# cso.write(b"x" * 2 ** 30)
# print(coutb.pending)

assert not coutb.pending
assert not cinb.pending
sso.do_handshake()
assert not coutb.pending
assert not cinb.pending



"""Trio - Pythonic async I/O for humans and snake people.
"""

# General layout:
#
# trio/_core/... is the self-contained core library. It does various
# shenanigans to export a consistent "core API", but parts of the core API are
# too low-level to be recommended for regular use.
#
# trio/*.py define a set of more usable tools on top of this. They import from
# trio._core and from each other.
#
# This file pulls together the friendly public API, by re-exporting the more
# innocuous bits of the _core API + the higher-level tools from trio/*.py.

from ._version import __version__

from ._core import (
    TrioInternalError, RunFinishedError, WouldBlock, Cancelled,
    BusyResourceError, ClosedResourceError, MultiError, run, open_nursery,
    open_cancel_scope, current_effective_deadline, TASK_STATUS_IGNORED,
    current_time, BrokenResourceError, EndOfChannel
)

from ._timeouts import (
    move_on_at, move_on_after, sleep_forever, sleep_until, sleep, fail_at,
    fail_after, TooSlowError
)

from ._sync import (
    Event, CapacityLimiter, Semaphore, Lock, StrictFIFOLock, Condition, Queue
)

from ._threads import (
    run_sync_in_worker_thread, current_default_worker_thread_limiter,
    BlockingTrioPortal
)

from ._highlevel_generic import aclose_forcefully, StapledStream

from ._channel import open_memory_channel

from ._signals import catch_signals, open_signal_receiver

from ._highlevel_socket import SocketStream, SocketListener

from ._file_io import open_file, wrap_file

from ._path import Path

from ._highlevel_serve_listeners import serve_listeners

from ._highlevel_open_tcp_stream import open_tcp_stream

from ._highlevel_open_tcp_listeners import open_tcp_listeners, serve_tcp

from ._highlevel_open_unix_stream import open_unix_socket

from ._highlevel_ssl_helpers import (
    open_ssl_over_tcp_stream, open_ssl_over_tcp_listeners, serve_ssl_over_tcp
)

from ._deprecate import TrioDeprecationWarning

# Imported by default
from . import hazmat
from . import socket
from . import abc
from . import ssl
# Not imported by default: testing
if False:
    from . import testing

_deprecate.enable_attribute_deprecations(__name__)
__deprecated_attributes__ = {
    "ClosedStreamError":
        _deprecate.DeprecatedAttribute(
            ClosedResourceError,
            "0.5.0",
            issue=36,
            instead=ClosedResourceError
        ),
    "ClosedListenerError":
        _deprecate.DeprecatedAttribute(
            ClosedResourceError,
            "0.5.0",
            issue=36,
            instead=ClosedResourceError
        ),
    "BrokenStreamError":
        _deprecate.DeprecatedAttribute(
            BrokenResourceError,
            "0.8.0",
            issue=620,
            instead=BrokenResourceError
        ),
    "ResourceBusyError":
        _deprecate.DeprecatedAttribute(
            BusyResourceError, "0.8.0", issue=620, instead=BusyResourceError
        ),
}

_deprecate.enable_attribute_deprecations(hazmat.__name__)

# Temporary hack to make sure _result is loaded, just during the deprecation
# period
from ._core import _result

hazmat.__deprecated_attributes__ = {
    "Result":
        _deprecate.DeprecatedAttribute(
            _core._result.Result,
            "0.5.0",
            issue=494,
            instead="outcome.Outcome"
        ),
    "Value":
        _deprecate.DeprecatedAttribute(
            _core._result.Value, "0.5.0", issue=494, instead="outcome.Value"
        ),
    "Error":
        _deprecate.DeprecatedAttribute(
            _core._result.Error, "0.5.0", issue=494, instead="outcome.Error"
        )
}

# Having the public path in .__module__ attributes is important for:
# - exception names in printed tracebacks
# - sphinx :show-inheritance:
# - deprecation warnings
# - pickle
# - probably other stuff
from ._util import fixup_module_metadata
fixup_module_metadata(__name__, globals())
fixup_module_metadata(hazmat.__name__, hazmat.__dict__)
fixup_module_metadata(socket.__name__, socket.__dict__)
fixup_module_metadata(abc.__name__, abc.__dict__)
fixup_module_metadata(ssl.__name__, ssl.__dict__)
del fixup_module_metadata



# If you want to use IPv6, then:
# - replace AF_INET with AF_INET6 everywhere
# - use the hostname "2.pool.ntp.org"
#   (see: https://news.ntppool.org/2011/06/continuing-ipv6-deployment/)

import trio
import struct
import datetime

def make_query_packet():
    """Construct a UDP packet suitable for querying an NTP server to ask for
    the current time."""

    # The structure of an NTP packet is described here:
    #   https://tools.ietf.org/html/rfc5905#page-19
    # They're always 48 bytes long, unless you're using extensions, which we
    # aren't.
    packet = bytearray(48)

    # The first byte contains 3 subfields:
    # first 2 bits: 11, leap second status unknown
    # next 3 bits: 100, NTP version indicator, 0b100 == 4 = version 4
    # last 3 bits: 011, NTP mode indicator, 0b011 == 3 == "client"
    packet[0] = 0b11100011

    # For an outgoing request, all other fields can be left as zeros.

    return packet

def extract_transmit_timestamp(ntp_packet):
    """Given an NTP packet, extract the "transmit timestamp" field, as a
    Python datetime."""

    # The transmit timestamp is the time that the server sent its response.
    # It's stored in bytes 40-47 of the NTP packet. See:
    #   https://tools.ietf.org/html/rfc5905#page-19
    encoded_transmit_timestamp = ntp_packet[40:48]

    # The timestamp is stored in the "NTP timestamp format", which is a 32
    # byte count of whole seconds, followed by a 32 byte count of fractions of
    # a second. See:
    #   https://tools.ietf.org/html/rfc5905#page-13
    seconds, fraction = struct.unpack("!II", encoded_transmit_timestamp)

    # The timestamp is the number of seconds since January 1, 1900 (ignoring
    # leap seconds). To convert it to a datetime object, we do some simple
    # datetime arithmetic:
    base_time = datetime.datetime(1900, 1, 1)
    offset = datetime.timedelta(seconds=seconds + fraction / 2**32)
    return base_time + offset

async def main():
    print("Our clock currently reads (in UTC):", datetime.datetime.utcnow())

    # Look up some random NTP servers.
    # (See www.pool.ntp.org for information about the NTP pool.)
    servers = await trio.socket.getaddrinfo(
        "pool.ntp.org",               # host
        "ntp",                        # port
        family=trio.socket.AF_INET,   # IPv4
        type=trio.socket.SOCK_DGRAM,  # UDP
    )

    # Construct an NTP query packet.
    query_packet = make_query_packet()

    # Create a UDP socket
    udp_sock = trio.socket.socket(
        family=trio.socket.AF_INET,   # IPv4
        type=trio.socket.SOCK_DGRAM,  # UDP
    )

    # Use the socket to send the query packet to each of the servers.
    print("-- Sending queries --")
    for server in servers:
        address = server[-1]
        print("Sending to:", address)
        await udp_sock.sendto(query_packet, address)

    # Read responses from the socket.
    print("-- Reading responses (for 10 seconds) --")
    with trio.move_on_after(10):
        while True:
            # We accept packets up to 1024 bytes long (though in practice NTP
            # packets will be much shorter).
            data, address = await udp_sock.recvfrom(1024)
            print("Got response from:", address)
            transmit_timestamp = extract_transmit_timestamp(data)
            print("Their clock read (in UTC):", transmit_timestamp)

trio.run(main)



from contextlib import contextmanager
import trio

class GracefulShutdownManager:
    def __init__(self):
        self._shutting_down = False
        self._cancel_scopes = set()

    def start_shutdown(self):
        self._shutting_down = True
        for cancel_scope in self._cancel_scopes:
            cancel_scope.cancel()

    @contextmanager
    def cancel_on_graceful_shutdown(self):
        with trio.open_cancel_scope() as cancel_scope:
            self._cancel_scopes.add(cancel_scope)
            if self._shutting_down:
                cancel_scope.cancel()
            try:
                yield
            finally:
                self._cancel_scopes.remove(cancel_scope)

    @property
    def shutting_down(self):
        return self._shutting_down

# Code can check gsm.shutting_down occasionally at appropriate points to see
# if it should exit.
#
# When doing operations that might block for an indefinite about of time and
# that should be aborted when a graceful shutdown starts, wrap them in 'with
# gsm.cancel_on_graceful_shutdown()'.
async def stream_handler(stream):
    while True:
        with gsm.cancel_on_graceful_shutdown():
            data = await stream.receive_some(...)
        if gsm.shutting_down:
            break

# To trigger the shutdown:
async def listen_for_shutdown_signals():
    with trio.catch_signals({signal.SIGINT, signal.SIGTERM}) as signal_aiter:
        async for batch in signal_aiter:
            gsm.start_shutdown()
            break
        # TODO: it'd be nice to have some logic like "if we get another
        # signal, or if 30 seconds pass, then do a hard shutdown".
        # That's easy enough:
        #
        # with trio.move_on_after(30):
        #     async for batch in signal_aiter:
        #         break
        # sys.exit()
        #
        # The trick is, if we do finish shutting down in (say) 10 seconds,
        # then we want to exit immediately. So I guess you'd need the main
        # part of the program to detect when it's finished shutting down, and
        # then cancel listen_for_shutdown_signals?
        #
        # I guess this would be a good place to use @smurfix's daemon task
        # construct:
        # https://github.com/python-trio/trio/issues/569#issuecomment-408419260



import time

# https://bitbucket.org/pypy/pypy/issues/2624/weird-performance-on-pypy3-when-reading
# COUNT = 100000
# f = open("/etc/passwd", "rt")
COUNT = 1000000
# With default buffering this test never even syscalls, and goes at about ~140
# ns per call, instead of ~500 ns/call for the syscall and related overhead.
# That's probably more fair -- the BufferedIOBase code can't service random
# accesses, even if your working set fits entirely in RAM.
f = open("/etc/passwd", "rb")#, buffering=0)

while True:
    start = time.perf_counter()
    for _ in range(COUNT):
        f.seek(0)
        f.read(1)
    between = time.perf_counter()
    for _ in range(COUNT):
        f.seek(0)
    end = time.perf_counter()

    both = (between - start) / COUNT * 1e9
    seek = (end - between) / COUNT * 1e9
    read = both - seek
    print("{:.2f} ns/(seek+read), {:.2f} ns/seek, estimate ~{:.2f} ns/read"
          .format(both, seek, read))



import os
import signal
import threading
import time
import socket
import select
import itertools

# Equivalent to the C function raise(), which Python doesn't wrap
if os.name == "nt":
    import cffi
    _ffi = cffi.FFI()
    _ffi.cdef("int raise(int);")
    _lib = _ffi.dlopen("api-ms-win-crt-runtime-l1-1-0.dll")
    signal_raise = getattr(_lib, "raise")
else:
    def signal_raise(signum):
        # Use pthread_kill to make sure we're actually using the wakeup fd on
        # Unix
        signal.pthread_kill(threading.get_ident(), signum)


def raise_SIGINT_soon():
    time.sleep(1)
    signal_raise(signal.SIGINT)
    # Sending 2 signals becomes reliable, as we'd expect (because we need
    # set-flags -> write-to-fd, and doing it twice does
    # write-to-fd -> set-flags -> write-to-fd -> set-flags)
    #signal_raise(signal.SIGINT)


def drain(sock):
    total = 0
    try:
        while True:
            total += len(sock.recv(1024))
    except BlockingIOError:
        pass
    return total


def main():
    writer, reader = socket.socketpair()
    writer.setblocking(False)
    reader.setblocking(False)

    signal.set_wakeup_fd(writer.fileno())

    # Keep trying until we lose the race...
    for attempt in itertools.count():
        print(f"Attempt {attempt}: start")

        # Make sure the socket is empty
        drained = drain(reader)
        if drained:
            print(f"Attempt {attempt}: ({drained} residual bytes discarded)")

        # Arrange for SIGINT to be delivered 1 second from now
        thread = threading.Thread(target=raise_SIGINT_soon)
        thread.start()

        # Fake an IO loop that's trying to sleep for 10 seconds (but will
        # hopefully get interrupted after just 1 second)
        start = time.perf_counter()
        target = start + 10
        try:
            select_calls = 0
            drained = 0
            while True:
                now = time.perf_counter()
                if now > target:
                    break
                select_calls += 1
                r, _, _ = select.select([reader], [], [], target - now)
                if r:
                    # In theory we should loop to fully drain the socket but
                    # honestly there's 1 byte in there at most and it'll be
                    # fine.
                    drained += drain(reader)
        except KeyboardInterrupt:
            pass
        else:
            print(f"Attempt {attempt}: no KeyboardInterrupt?!")

        # We expect a successful run to take 1 second, and a failed run to
        # take 10 seconds, so 2 seconds is a reasonable cutoff to distinguish
        # them.
        duration = time.perf_counter() - start
        if duration < 2:
            print(f"Attempt {attempt}: OK, trying again "
                  f"(select_calls = {select_calls}, drained = {drained})")
        else:
            print(f"Attempt {attempt}: FAILED, took {duration} seconds")
            print(f"select_calls = {select_calls}, drained = {drained}")
            break

        thread.join()

if __name__ == "__main__":
    main()



"""
This namespace represents low-level functionality not intended for daily use,
but useful for extending Trio's functionality.
"""

import sys

# This is the union of a subset of trio/_core/ and some things from trio/*.py.
# See comments in trio/__init__.py for details. To make static analysis easier,
# this lists all possible symbols from trio._core, and then we prune those that
# aren't available on this system. After that we add some symbols from trio/*.py.

__all__ = [
    "cancel_shielded_checkpoint",
    "Abort",
    "wait_task_rescheduled",
    "enable_ki_protection",
    "disable_ki_protection",
    "currently_ki_protected",
    "Task",
    "checkpoint",
    "current_task",
    "current_root_task",
    "checkpoint_if_cancelled",
    "spawn_system_task",
    "reschedule",
    "remove_instrument",
    "add_instrument",
    "current_clock",
    "current_statistics",
    "ParkingLot",
    "UnboundedQueue",
    "RunVar",
    "wait_writable",
    "wait_readable",
    "notify_fd_close",
    "wait_socket_readable",
    "wait_socket_writable",
    "notify_socket_close",
    "TrioToken",
    "current_trio_token",
    # kqueue symbols
    "current_kqueue",
    "monitor_kevent",
    "wait_kevent",
    # windows symbols
    "current_iocp",
    "register_with_iocp",
    "wait_overlapped",
    "monitor_completion_key",
]

from . import _core
# Some hazmat symbols are platform specific
for _sym in list(__all__):
    if hasattr(_core, _sym):
        globals()[_sym] = getattr(_core, _sym)
    else:
        # Fool static analysis (at least PyCharm's) into thinking that we're
        # not modifying __all__, so it can trust the static list up above.
        # https://github.com/python-trio/trio/pull/316#issuecomment-328255867
        # This was useful in September 2017. If it's not September 2017 then
        # who knows.
        remove_from_all = __all__.remove
        remove_from_all(_sym)

# Import bits from trio/*.py
if sys.platform.startswith("win"):
    from ._wait_for_object import WaitForSingleObject
    __all__ += ["WaitForSingleObject"]



"""
Demo/exploration of how to print a task tree. Outputs:

<init>
├─ __main__.main
│  ├─ __main__.child1
│  │  ├─ trio.sleep_forever
│  │  ├─ __main__.child2
│  │  │  ├─ trio.sleep_forever
│  │  │  └─ trio.sleep_forever
│  │  └─ __main__.child2
│  │     ├─ trio.sleep_forever
│  │     └─ trio.sleep_forever
│  └─ (nested nursery)
│     └─ __main__.child1
│        ├─ trio.sleep_forever
│        ├─ __main__.child2
│        │  ├─ trio.sleep_forever
│        │  └─ trio.sleep_forever
│        └─ __main__.child2
│           ├─ trio.sleep_forever
│           └─ trio.sleep_forever
└─ <call soon task>

"""

import trio
import trio.testing

MID_PREFIX = "├─ "
MID_CONTINUE = "│  "
END_PREFIX = "└─ "
END_CONTINUE = " " * len(END_PREFIX)


def current_root_task():
    task = trio.hazmat.current_task()
    while task.parent_nursery is not None:
        task = task.parent_nursery.parent_task
    return task


def _render_subtree(name, rendered_children):
    lines = []
    lines.append(name)
    for child_lines in rendered_children:
        if child_lines is rendered_children[-1]:
            first_prefix = END_PREFIX
            rest_prefix = END_CONTINUE
        else:
            first_prefix = MID_PREFIX
            rest_prefix = MID_CONTINUE
        lines.append(first_prefix + child_lines[0])
        for child_line in child_lines[1:]:
            lines.append(rest_prefix + child_line)
    return lines


def _rendered_nursery_children(nursery):
    return [task_tree_lines(t) for t in nursery.child_tasks]


def task_tree_lines(task=None):
    if task is None:
        task = current_root_task()
    rendered_children = []
    nurseries = list(task.child_nurseries)
    while nurseries:
        nursery = nurseries.pop()
        nursery_children = _rendered_nursery_children(nursery)
        if rendered_children:
            nested = _render_subtree("(nested nursery)", rendered_children)
            nursery_children.append(nested)
        rendered_children = nursery_children
    return _render_subtree(task.name, rendered_children)


def print_task_tree(task=None):
    for line in task_tree_lines(task):
        print(line)


################################################################


async def child2():
    async with trio.open_nursery() as nursery:
        nursery.start_soon(trio.sleep_forever)
        nursery.start_soon(trio.sleep_forever)


async def child1():
    async with trio.open_nursery() as nursery:
        nursery.start_soon(child2)
        nursery.start_soon(child2)
        nursery.start_soon(trio.sleep_forever)


async def main():
    async with trio.open_nursery() as nursery0:
        nursery0.start_soon(child1)
        async with trio.open_nursery() as nursery1:
            nursery1.start_soon(child1)

            await trio.testing.wait_all_tasks_blocked()
            print_task_tree()
            nursery0.cancel_scope.cancel()


trio.run(main)



# This demonstrates a PyPy bug:
#   https://bitbucket.org/pypy/pypy/issues/2578/

import socket
import ssl
import threading
import os

#client_sock, server_sock = socket.socketpair()
listen_sock = socket.socket()
listen_sock.bind(("127.0.0.1", 0))
listen_sock.listen(1)
client_sock = socket.socket()
client_sock.connect(listen_sock.getsockname())
server_sock, _ = listen_sock.accept()

server_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
server_ctx.load_cert_chain("trio-test-1.pem")
server = server_ctx.wrap_socket(
    server_sock,
    server_side=True,
    suppress_ragged_eofs=False,
    do_handshake_on_connect=False,
)

client_ctx = ssl.create_default_context(cafile="trio-test-CA.pem")
client = client_ctx.wrap_socket(
    client_sock,
    server_hostname="trio-test-1.example.org",
    suppress_ragged_eofs=False,
    do_handshake_on_connect=False,
)

server_handshake_thread = threading.Thread(target=server.do_handshake)
server_handshake_thread.start()
client_handshake_thread = threading.Thread(target=client.do_handshake)
client_handshake_thread.start()

server_handshake_thread.join()
client_handshake_thread.join()

# Now we have two SSLSockets that have established an encrypted connection
# with each other

assert client.getpeercert() is not None
client.sendall(b"x")
assert server.recv(10) == b"x"

# A few different ways to make attempts to read/write the socket's fd return
# weird failures at the operating system level

# Attempting to send on a socket after shutdown should raise EPIPE or similar
server.shutdown(socket.SHUT_WR)

# Attempting to read/write to the fd after it's closed should raise EBADF
#os.close(server.fileno())

# Attempting to read/write to an fd opened with O_DIRECT raises EINVAL in most
# cases (unless you're very careful with alignment etc. which openssl isn't)
#os.dup2(os.open("/tmp/blah-example-file", os.O_RDWR | os.O_CREAT | os.O_DIRECT), server.fileno())

# Sending or receiving
server.sendall(b"hello")
#server.recv(10)



import trio

async def run_test(nominal_backlog):
    print("--\nnominal:", nominal_backlog)

    listen_sock = trio.socket.socket()
    await listen_sock.bind(("127.0.0.1", 0))
    listen_sock.listen(nominal_backlog)
    client_socks = []
    while True:
        client_sock = trio.socket.socket()
        # Generally the response to the listen buffer being full is that the
        # SYN gets dropped, and the client retries after 1 second. So we
        # assume that any connect() call to localhost that takes >0.5 seconds
        # indicates a dropped SYN.
        with trio.move_on_after(0.5) as cancel_scope:
            await client_sock.connect(listen_sock.getsockname())
        if cancel_scope.cancelled_caught:
            break
        client_socks.append(client_sock)
    print("actual:", len(client_socks))
    for client_sock in client_socks:
        client_sock.close()

for nominal_backlog in [10, trio.socket.SOMAXCONN, 65535]:
    trio.run(run_test, nominal_backlog)



# Estimate the cost of simply passing some data into a thread and back, in as
# minimal a fashion as possible.
#
# This is useful to get a sense of the *lower-bound* cost of
# run_sync_in_worker_thread

import threading
from queue import Queue
import time

COUNT = 10000

def worker(in_q, out_q):
    while True:
        job = in_q.get()
        out_q.put(job())

def main():
    in_q = Queue()
    out_q = Queue()

    t = threading.Thread(target=worker, args=(in_q, out_q))
    t.start()

    while True:
        start = time.monotonic()
        for _ in range(COUNT):
            in_q.put(lambda: None)
            out_q.get()
        end = time.monotonic()
        print("{:.2f} µs/job".format((end - start) / COUNT * 1e6))

main()



# Scenario:
# - TLS connection is set up successfully
# - client sends close_notify then closes socket
# - server receives the close_notify then attempts to send close_notify back
#
# On CPython, the last step raises BrokenPipeError. On PyPy, it raises
# SSLEOFError.
#
# SSLEOFError seems a bit perverse given that it's supposed to mean "EOF
# occurred in violation of protocol", and the client's behavior here is
# explicitly allowed by the RFCs. But maybe openssl is just perverse like
# that, and it's a coincidence that CPython and PyPy act differently here? I
# don't know if this is a bug or not.
#
# (Using: debian's CPython 3.5 or 3.6, and pypy3 5.8.0-beta)

import socket
import ssl
import threading

client_sock, server_sock = socket.socketpair()

client_done = threading.Event()

def server_thread_fn():
    server_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    server_ctx.load_cert_chain("trio-test-1.pem")
    server = server_ctx.wrap_socket(
        server_sock,
        server_side=True,
        suppress_ragged_eofs=False,
    )
    while True:
        data = server.recv(4096)
        print("server got:", data)
        if not data:
            print("server waiting for client to finish everything")
            client_done.wait()
            print("server attempting to send back close-notify")
            server.unwrap()
            print("server ok")
            break
        server.sendall(data)

server_thread = threading.Thread(target=server_thread_fn)
server_thread.start()

client_ctx = ssl.create_default_context(cafile="trio-test-CA.pem")
client = client_ctx.wrap_socket(
    client_sock,
    server_hostname="trio-test-1.example.org")


# Now we have two SSLSockets that have established an encrypted connection
# with each other

assert client.getpeercert() is not None
client.sendall(b"x")
assert client.recv(10) == b"x"

# The client sends a close-notify, and then immediately closes the connection
# (as explicitly permitted by the TLS RFCs).

# This is a slightly odd construction, but if you trace through the ssl module
# far enough you'll see that it's equivalent to calling SSL_shutdown() once,
# which generates the close_notify, and then immediately calling it again,
# which checks for the close_notify and then immediately raises
# SSLWantReadError because of course it hasn't arrived yet:
print("client sending close_notify")
client.setblocking(False)
try:
    client.unwrap()
except ssl.SSLWantReadError:
    print("client got SSLWantReadError as expected")
else:
    assert False
client.close()
client_done.set()



import trio
import trio.testing

import jedi
import os
import pytest
import sys

from .. import _core


def test_core_is_properly_reexported():
    # Each export from _core should be re-exported by exactly one of these
    # three modules:
    sources = [trio, trio.hazmat, trio.testing]
    for symbol in dir(_core):
        if symbol.startswith('_') or symbol == 'tests':
            continue
        found = 0
        for source in sources:
            if (
                symbol in dir(source)
                and getattr(source, symbol) is getattr(_core, symbol)
            ):
                found += 1
        print(symbol, found)
        assert found == 1


@pytest.mark.skipif(
    sys.version_info.releaselevel == "alpha",
    reason="skip pylint on in-development Python",
)
def test_pylint_sees_all_non_underscore_symbols_in_namespace():
    # Test pylints ast to contain the same content as dir(trio)
    from pylint.lint import PyLinter
    linter = PyLinter()
    ast_set = set(linter.get_ast(trio.__file__, 'trio'))
    trio_set = set([symbol for symbol in dir(trio) if symbol[0] != '_'])
    trio_set.remove('tests')
    assert trio_set - ast_set == set([])


def test_jedi_sees_all_completions():
    # Test the jedi completion library get all in dir(trio)
    try:
        script = jedi.Script("import trio; trio.")
        completions = script.completions()
        trio_set = set([symbol for symbol in dir(trio) if symbol[:2] != '__'])
        jedi_set = set([cmp.name for cmp in completions])
        assert trio_set - jedi_set == set([])
    except NotImplementedError:
        pytest.skip("jedi does not yet support {}".format(sys.version))



# On windows, what does SO_EXCLUSIVEADDRUSE actually do? Apparently not what
# the documentation says!
# See: https://stackoverflow.com/questions/45624916/
#
# Specifically, this script seems to demonstrate that it only creates
# conflicts between listening sockets, *not* lingering connected sockets.

import socket
from contextlib import contextmanager

@contextmanager
def report_outcome(tagline):
    try:
        yield
    except OSError as exc:
        print("{}: failed".format(tagline))
        print("    details: {!r}".format(exc))
    else:
        print("{}: succeeded".format(tagline))

# Set up initial listening socket
lsock = socket.socket()
lsock.setsockopt(socket.SOL_SOCKET, socket.SO_EXCLUSIVEADDRUSE, 1)
lsock.bind(("127.0.0.1", 0))
sockaddr = lsock.getsockname()
lsock.listen(10)

# Make connected client and server sockets
csock = socket.socket()
csock.connect(sockaddr)
ssock, _ = lsock.accept()

print("lsock", lsock.getsockname())
print("ssock", ssock.getsockname())

# Can't make a second listener while the first exists
probe = socket.socket()
probe.setsockopt(socket.SOL_SOCKET, socket.SO_EXCLUSIVEADDRUSE, 1)
with report_outcome("rebind with existing listening socket"):
    probe.bind(sockaddr)

# Now we close the first listen socket, while leaving the connected sockets
# open:
lsock.close()
# This time binding succeeds (contra MSDN!)
probe = socket.socket()
probe.setsockopt(socket.SOL_SOCKET, socket.SO_EXCLUSIVEADDRUSE, 1)
with report_outcome("rebind with live connected sockets"):
    probe.bind(sockaddr)
    probe.listen(10)
    print("probe", probe.getsockname())
    print("ssock", ssock.getsockname())
probe.close()

# Server-initiated close to trigger TIME_WAIT status
ssock.send(b"x")
assert csock.recv(1) == b"x"
ssock.close()
assert csock.recv(1) == b""

# And does the TIME_WAIT sock prevent binding?
probe = socket.socket()
probe.setsockopt(socket.SOL_SOCKET, socket.SO_EXCLUSIVEADDRUSE, 1)
with report_outcome("rebind with TIME_WAIT socket"):
    probe.bind(sockaddr)
    probe.listen(10)
probe.close()



import sys
import trio

sys.stderr = sys.stdout

async def child1():
    raise ValueError

async def child2():
    async with trio.open_nursery() as nursery:
        nursery.start_soon(grandchild1)
        nursery.start_soon(grandchild2)

async def grandchild1():
    raise KeyError

async def grandchild2():
    raise NameError("Bob")

async def main():
    async with trio.open_nursery() as nursery:
        nursery.start_soon(child1)
        nursery.start_soon(child2)
        #nursery.start_soon(grandchild1)

trio.run(main)



import ssl
import socket
import threading
from contextlib import contextmanager

BUFSIZE = 4096

server_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
server_ctx.load_cert_chain("trio-test-1.pem")

def _ssl_echo_serve_sync(sock):
    try:
        wrapped = server_ctx.wrap_socket(sock, server_side=True)
        while True:
            data = wrapped.recv(BUFSIZE)
            if not data:
                wrapped.unwrap()
                return
            wrapped.sendall(data)
    except BrokenPipeError:
        pass

@contextmanager
def echo_server_connection():
    client_sock, server_sock = socket.socketpair()
    with client_sock, server_sock:
        t = threading.Thread(
            target=_ssl_echo_serve_sync, args=(server_sock,), daemon=True)
        t.start()

        yield client_sock

class ManuallyWrappedSocket:
    def __init__(self, ctx, sock, **kwargs):
        self.incoming = ssl.MemoryBIO()
        self.outgoing = ssl.MemoryBIO()
        self.obj = ctx.wrap_bio(self.incoming, self.outgoing, **kwargs)
        self.sock = sock

    def _retry(self, fn, *args):
        finished = False
        while not finished:
            want_read = False
            try:
                ret = fn(*args)
            except ssl.SSLWantReadError:
                want_read = True
            except ssl.SSLWantWriteError:
                # can't happen, but if it did this would be the right way to
                # handle it anyway
                pass
            else:
                finished = True
            # do any sending
            data = self.outgoing.read()
            if data:
                self.sock.sendall(data)
            # do any receiving
            if want_read:
                data = self.sock.recv(BUFSIZE)
                if not data:
                    self.incoming.write_eof()
                else:
                    self.incoming.write(data)
            # then retry if necessary
        return ret

    def do_handshake(self):
        self._retry(self.obj.do_handshake)

    def recv(self, bufsize):
        return self._retry(self.obj.read, bufsize)

    def sendall(self, data):
        self._retry(self.obj.write, data)

    def unwrap(self):
        self._retry(self.obj.unwrap)
        return self.sock


def wrap_socket_via_wrap_socket(ctx, sock, **kwargs):
    return ctx.wrap_socket(sock, do_handshake_on_connect=False, **kwargs)

def wrap_socket_via_wrap_bio(ctx, sock, **kwargs):
    return ManuallyWrappedSocket(ctx, sock, **kwargs)


for wrap_socket in [
        wrap_socket_via_wrap_socket,
        wrap_socket_via_wrap_bio,
]:
    print("\n--- checking {} ---\n".format(wrap_socket.__name__))

    print("checking with do_handshake + correct hostname...")
    with echo_server_connection() as client_sock:
        client_ctx = ssl.create_default_context(cafile="trio-test-CA.pem")
        wrapped = wrap_socket(
            client_ctx, client_sock, server_hostname="trio-test-1.example.org")
        wrapped.do_handshake()
        wrapped.sendall(b"x")
        assert wrapped.recv(1) == b"x"
        wrapped.unwrap()
    print("...success")

    print("checking with do_handshake + wrong hostname...")
    with echo_server_connection() as client_sock:
        client_ctx = ssl.create_default_context(cafile="trio-test-CA.pem")
        wrapped = wrap_socket(
            client_ctx, client_sock, server_hostname="trio-test-2.example.org")
        try:
            wrapped.do_handshake()
        except Exception:
            print("...got error as expected")
        else:
            print("??? no error ???")

    print("checking withOUT do_handshake + wrong hostname...")
    with echo_server_connection() as client_sock:
        client_ctx = ssl.create_default_context(cafile="trio-test-CA.pem")
        wrapped = wrap_socket(
            client_ctx, client_sock, server_hostname="trio-test-2.example.org")
        # We forgot to call do_handshake
        # But the hostname is wrong so something had better error out...
        sent = b"x"
        print("sending", sent)
        wrapped.sendall(sent)
        got = wrapped.recv(1)
        print("got:", got)
        assert got == sent
        print("!!!! successful chat with invalid host! we have been haxored!")



import sys
import trio

(COUNT_STR,) = sys.argv[1:]
COUNT = int(COUNT_STR)

async def main():
    async with trio.open_nursery() as nursery:
        for _ in range(COUNT):
            nursery.start_soon(trio.sleep, 1)

trio.run(main)



import trio
import time

async def loopy():
    try:
        while True:
            time.sleep(0.01)
            await trio.sleep(0)
    except KeyboardInterrupt:
        print("KI!")

async def main():
    async with trio.open_nursery() as nursery:
        nursery.start_soon(loopy)
        nursery.start_soon(loopy)
        nursery.start_soon(loopy)

trio.run(main)



import trio
import os
import socket
import errno

bad_socket = socket.socket()

class BlockingReadTimeoutError(Exception):
    pass

async def blocking_read_with_timeout(fd, count, timeout):
    print("reading from fd", fd)
    cancel_requested = False

    async def kill_it_after_timeout(new_fd):
        print("sleeping")
        await trio.sleep(timeout)
        print("breaking the fd")
        os.dup2(bad_socket.fileno(), new_fd, inheritable=False)
        # MAGIC
        print("setuid(getuid())")
        os.setuid(os.getuid())
        nonlocal cancel_requested
        cancel_requested = True

    new_fd = os.dup(fd)
    print("working fd is", new_fd)
    try:
        async with trio.open_nursery() as nursery:
            nursery.start_soon(kill_it_after_timeout, new_fd)
            try:
                data = await trio.run_sync_in_worker_thread(os.read, new_fd, count)
            except OSError as exc:
                if cancel_requested and exc.errno == errno.ENOTCONN:
                    # Call was successfully cancelled. In a real version we'd
                    # integrate properly with trio's cancellation tools; here
                    # we'll just raise an arbitrary error.
                    raise BlockingReadTimeoutError from None
            print("got", data)
            nursery.cancel_scope.cancel()
            return data
    finally:
        os.close(new_fd)

trio.run(blocking_read_with_timeout, 0, 10, 2)



# This is some very messy notes on how we might implement a thread cache

import threading
import Queue

# idea:
#
# unbounded thread pool; tracks how many threads are "available" and how much
# work there is to do; if work > available threads, spawn a new thread
#
# if a thread sits idle for >N ms, exit
#
# we don't need to support job cancellation
#
# we do need to mark a thread as "available" just before it
# signals back to trio that it's done, to maintain the invariant that all
# unavailable threads are inside the limiter= protection
#
# maintaining this invariant while exiting can be a bit tricky
#
# maybe a simple target should be to always have 1 idle thread

# XX we can't use a single shared dispatch queue, because we need LIFO
# scheduling, or else the idle-thread timeout won't work!
#
# instead, keep a list/deque/OrderedDict/something of idle threads, and
# dispatch by popping one off; put things back by pushing them on the end
# maybe one shared dispatch Lock, plus a Condition for each thread
# dispatch by dropping the job into the place where the thread can see it and
# then signalling its Condition? or could have separate locks

@attr.s(frozen=True)
class Job:
    main = attr.ib()
    main_args = attr.ib()
    finish = attr.ib()
    finish_args = attr.ib()

class EXIT:
    pass

class ThreadCache:
    def __init__(self):
        self._lock = threading.Lock()
        self._idle_workers = deque()
        self._closed = False

    def close(self):
        self._closed = True
        with self._lock:
            while self._idle_workers:
                self._idle_workers.pop().submit(None)

    def submit(self, job):
        with self._lock:
            if not self._idle_workers:
                WorkerThread(self, self._lock, job)
            else:
                worker = self._idle_workers.pop()
                worker.submit(job)

    # Called from another thread
    # Must be called with the lock held
    def remove_idle_worker(self, worker):
        self._idle_workers.remove(worker)

    # Called from another thread
    # Lock is *not* held
    def add_idle_worker(self, worker):
        if self._closed:
            with self._lock:
            worker.submit
        self._idle_workers.append(worker)

# XX thread name

IDLE_TIMEOUT = 1.0

class WorkerThread:
    def __init__(self, cache, lock, initial_job):
        self._cache = cache
        self._condition = threading.Condition(lock)
        self._job = None
        self._thread = threading.Thread(
            target=self._loop, args=(initial_job,), daemon=True)
        self._thread.start()

    # Must be called with the lock held
    def submit(self, job):
        assert self._job is None
        self._job = job
        self._condition.notify()

    def _loop(self, initial_job):
        self._run_job(initial_job)
        while True:
            with self._condition:
                self._condition.wait(IDLE_TIMEOUT):
                job = self._job
                self._job = None
                if job is None:
                    self._cache.remove_idle_worker(self)
                    return
            # Dropped the lock, and have a job to do
            self._run_job(job)

    def _run_job(self, job):
        job.main(*job.main_args)
        self._cache.add_idle_worker(self)
        job.finish(*job.finish_args)


# Probably the interface should be: trio.hazmat.call_soon_in_worker_thread?

# Enqueueing work:
#   put into unbounded queue
#   with lock:
#       if idle_threads:
#           idle_threads -= 1
#       else:
#           spawn a new thread (it starts out non-idle)
#
# Thread shutdown:
#   with lock:
#       idle_threads -= 1
#       check for work one last time, and then either exit or do it
#
# Thread startup:
#
# check for work
# while True:
#   mark self as idle
#   check for work (with timeout)
#   either do work or shutdown

# if we want to support QueueUserAPC cancellation, we need a way to get back
# the thread id... maybe that just works like
#
# def WaitForSingleObjectEx_thread_fn(...):
#   with lock:
#     check if already cancelled
#     put our thread id where main thread can find it
#   WaitForSingleObjectEx(...)



import textwrap
import io
import time

methods = {"fileno"}

class Proxy1:
    strategy = "__getattr__"
    works_for = "any attr"

    def __init__(self, wrapped):
        self._wrapped = wrapped

    def __getattr__(self, name):
        if name in methods:
            return getattr(self._wrapped, name)
        raise AttributeError(name)

################################################################

class Proxy2:
    strategy = "generated methods (getattr + closure)"
    works_for = "methods"

    def __init__(self, wrapped):
        self._wrapped = wrapped

def add_wrapper(cls, method):
    def wrapper(self, *args, **kwargs):
        return getattr(self._wrapped, method)(*args, **kwargs)
    setattr(cls, method, wrapper)

for method in methods:
    add_wrapper(Proxy2, method)

################################################################

class Proxy3:
    strategy = "generated methods (exec)"
    works_for = "methods"

    def __init__(self, wrapped):
        self._wrapped = wrapped

def add_wrapper(cls, method):
    code = textwrap.dedent("""
        def wrapper(self, *args, **kwargs):
            return self._wrapped.{}(*args, **kwargs)
    """.format(method))
    ns = {}
    exec(code, ns)
    setattr(cls, method, ns["wrapper"])

for method in methods:
    add_wrapper(Proxy3, method)

################################################################

class Proxy4:
    strategy = "generated properties (getattr + closure)"
    works_for = "any attr"

    def __init__(self, wrapped):
        self._wrapped = wrapped

def add_wrapper(cls, attr):
    def getter(self):
        return getattr(self._wrapped, attr)

    def setter(self, newval):
        setattr(self._wrapped, attr, newval)

    def deleter(self):
        delattr(self._wrapped, attr)

    setattr(cls, attr, property(getter, setter, deleter))

for method in methods:
    add_wrapper(Proxy4, method)

################################################################

class Proxy5:
    strategy = "generated properties (exec)"
    works_for = "any attr"

    def __init__(self, wrapped):
        self._wrapped = wrapped

def add_wrapper(cls, attr):
    code = textwrap.dedent("""
        def getter(self):
            return self._wrapped.{attr}

        def setter(self, newval):
            self._wrapped.{attr} = newval

        def deleter(self):
            del self._wrapped.{attr}
    """.format(attr=attr))
    ns = {}
    exec(code, ns)
    setattr(cls, attr, property(ns["getter"], ns["setter"], ns["deleter"]))

for method in methods:
    add_wrapper(Proxy5, method)

################################################################

# methods only
class Proxy6:
    strategy = "copy attrs from wrappee to wrapper"
    works_for = "methods + constant attrs"

    def __init__(self, wrapper):
        self._wrapper = wrapper

        for method in methods:
            setattr(self, method, getattr(self._wrapper, method))
    

################################################################

classes = [Proxy1, Proxy2, Proxy3, Proxy4, Proxy5, Proxy6]

def check(cls):
    with open("/etc/passwd") as f:
        p = cls(f)
        assert p.fileno() == f.fileno()

for cls in classes:
    check(cls)

f = open("/etc/passwd")
objs = [c(f) for c in classes]

COUNT = 1000000
try:
    import __pypy__
except ImportError:
    pass
else:
    COUNT *= 10

while True:
    print("-------")
    for obj in objs:
        start = time.perf_counter()
        for _ in range(COUNT):
            obj.fileno()
            #obj.fileno
        end = time.perf_counter()
        per_usec = COUNT / (end - start) / 1e6
        print("{:7.2f} / us: {} ({})"
              .format(per_usec, obj.strategy, obj.works_for))



import sys

import os
import pytest
import signal

from ... import _core
from ..._subprocess.linux_waitpid import waitpid

pytestmark = pytest.mark.skipif(
    sys.platform != "linux", reason="linux waitpid only works on linux"
)


async def test_waitpid():
    pid = os.spawnvp(os.P_NOWAIT, "/bin/false", ("false",))
    result = await waitpid(pid)
    # exit code is a 16-bit int: (code, signal)
    assert result[0] == pid
    assert os.WIFEXITED(result[1]) and os.WEXITSTATUS(result[1]) == 1

    pid2 = os.spawnvp(os.P_NOWAIT, "/bin/true", ("true",))
    result = await waitpid(pid2)
    assert result[0] == pid2
    assert os.WIFEXITED(result[1]) and os.WEXITSTATUS(result[1]) == 0

    pid3 = os.spawnvp(os.P_NOWAIT, "/bin/sleep", ("/bin/sleep", "5"))
    os.kill(pid3, signal.SIGKILL)
    result = await waitpid(pid3)
    assert result[0] == pid3
    status = os.WTERMSIG(result[1])
    assert os.WIFSIGNALED(result[1]) and status == 9


async def test_waitpid_no_process():
    with pytest.raises(ChildProcessError):
        # this PID does exist, but it's ourselves
        # which doesn't work
        await waitpid(os.getpid())



from types import CodeType

# Has to be a string :-(
sentinel = "_unique_name"

def f():
    print(locals())

# code(argcount, kwonlyargcount, nlocals, stacksize, flags, codestring,
#       constants, names, varnames, filename, name, firstlineno,
#       lnotab[, freevars[, cellvars]])
new_code = CodeType(
    f.__code__.co_argcount,
    f.__code__.co_kwonlyargcount + 1,
    f.__code__.co_nlocals + 1,
    f.__code__.co_stacksize,
    f.__code__.co_flags,
    f.__code__.co_code,
    f.__code__.co_consts,
    f.__code__.co_names,
    f.__code__.co_varnames + (sentinel,),
    f.__code__.co_filename,
    f.__code__.co_name,
    f.__code__.co_firstlineno,
    f.__code__.co_lnotab,
    f.__code__.co_freevars,
    f.__code__.co_cellvars,
)

f.__code__ = new_code
f.__kwdefaults__ = {sentinel: "fdsa"}

f()



import fcntl
import os
from typing import Tuple

from .. import _core, BrokenResourceError
from .._abc import SendStream, ReceiveStream

__all__ = ["PipeSendStream", "PipeReceiveStream", "make_pipe"]


class _PipeMixin:
    def __init__(self, pipefd: int) -> None:
        if not isinstance(pipefd, int):
            raise TypeError(
                "{0.__class__.__name__} needs a pipe fd".format(self)
            )

        self._pipe = pipefd
        self._closed = False

        flags = fcntl.fcntl(self._pipe, fcntl.F_GETFL)
        fcntl.fcntl(self._pipe, fcntl.F_SETFL, flags | os.O_NONBLOCK)

    def _close(self):
        if self._closed:
            return

        self._closed = True
        os.close(self._pipe)

    async def aclose(self):
        # XX: This would be in _close, but this can only be used from an
        # async context.
        _core.notify_fd_close(self._pipe)
        self._close()
        await _core.checkpoint()

    def fileno(self) -> int:
        """Gets the file descriptor for this pipe."""
        return self._pipe

    def __del__(self):
        self._close()


class PipeSendStream(_PipeMixin, SendStream):
    """Represents a send stream over an os.pipe object."""

    async def send_all(self, data: bytes):
        # we have to do this no matter what
        await _core.checkpoint()
        if self._closed:
            raise _core.ClosedResourceError("this pipe is already closed")

        if not data:
            return

        length = len(data)
        # adapted from the SocketStream code
        with memoryview(data) as view:
            total_sent = 0
            while total_sent < length:
                with view[total_sent:] as remaining:
                    try:
                        total_sent += os.write(self._pipe, remaining)
                    except BrokenPipeError as e:
                        await _core.checkpoint()
                        raise BrokenResourceError from e
                    except BlockingIOError:
                        await self.wait_send_all_might_not_block()

    async def wait_send_all_might_not_block(self) -> None:
        if self._closed:
            await _core.checkpoint()
            raise _core.ClosedResourceError("This pipe is already closed")

        try:
            await _core.wait_writable(self._pipe)
        except BrokenPipeError as e:
            # kqueue: raises EPIPE on wait_writable instead
            # of sending, which is annoying
            # also doesn't checkpoint so we have to do that
            # ourselves here too
            await _core.checkpoint()
            raise BrokenResourceError from e


class PipeReceiveStream(_PipeMixin, ReceiveStream):
    """Represents a receive stream over an os.pipe object."""

    async def receive_some(self, max_bytes: int) -> bytes:
        if self._closed:
            await _core.checkpoint()
            raise _core.ClosedResourceError("this pipe is already closed")

        if not isinstance(max_bytes, int):
            await _core.checkpoint()
            raise TypeError("max_bytes must be integer >= 1")

        if max_bytes < 1:
            await _core.checkpoint()
            raise ValueError("max_bytes must be integer >= 1")

        while True:
            try:
                await _core.checkpoint_if_cancelled()
                data = os.read(self._pipe, max_bytes)
            except BlockingIOError:
                await _core.wait_readable(self._pipe)
            else:
                await _core.cancel_shielded_checkpoint()
                break

        return data


async def make_pipe() -> Tuple[PipeSendStream, PipeReceiveStream]:
    """Makes a new pair of pipes."""
    (r, w) = os.pipe()
    return PipeSendStream(w), PipeReceiveStream(r)



# Little script to get a rough estimate of how much memory each task takes

import resource
import trio
import trio.testing

LOW = 1000
HIGH = 10000

async def tinytask():
    await trio.sleep_forever()

async def measure(count):
    async with trio.open_nursery() as nursery:
        for _ in range(count):
            nursery.start_soon(tinytask)
        await trio.testing.wait_all_tasks_blocked()
        nursery.cancel_scope.cancel()
        return resource.getrusage(resource.RUSAGE_SELF)


async def main():
    low_usage = await measure(LOW)
    high_usage = await measure(HIGH + LOW)

    print("Memory usage per task:",
          (high_usage.ru_maxrss - low_usage.ru_maxrss) / HIGH)
    print("(kilobytes on Linux, bytes on macOS)")

trio.run(main)



# This is a reproducer for:
#   https://bugs.python.org/issue30744
#   https://bitbucket.org/pypy/pypy/issues/2591/

import sys
import threading
import time

COUNT = 100

def slow_tracefunc(frame, event, arg):
    # A no-op trace function that sleeps briefly to make us more likely to hit
    # the race condition.
    time.sleep(0.01)
    return slow_tracefunc

def run_with_slow_tracefunc(fn):
    # settrace() only takes effect when you enter a new frame, so we need this
    # little dance:
    sys.settrace(slow_tracefunc)
    return fn()

def outer():
    x = 0
    # We hide the done variable inside a list, because we want to use it to
    # communicate between the main thread and the looper thread, and the bug
    # here is that variable assignments made in the main thread disappear
    # before the child thread can see them...
    done = [False]

    def traced_looper():
        # Force w_locals to be instantiated (only matters on PyPy; on CPython
        # you can comment this line out and everything stays the same)
        print(locals())
        # Force x to be closed over (could use 'nonlocal' on py3)
        if False:
            x
        # Random nonsense whose only purpose is to trigger lots of calls to
        # the trace func
        count = 0
        while not done[0]:
            count += 1
        return count

    t = threading.Thread(target=run_with_slow_tracefunc, args=(traced_looper,))
    t.start()

    for i in range(COUNT):
        print("after {} increments, x is {}".format(i, x))
        x += 1
        time.sleep(0.01)

    done[0] = True
    t.join()

    print("Final discrepancy: {} (should be 0)".format(COUNT - x))

outer()



import trio
import time

LOOPS = 0
RUNNING = True

async def reschedule_loop(depth):
    if depth == 0:
        global LOOPS
        while RUNNING:
            LOOPS += 1
            await trio.sleep(0)
            #await trio.hazmat.cancel_shielded_checkpoint()
    else:
        await reschedule_loop(depth - 1)

async def report_loop():
    global RUNNING
    try:
        while True:
            start_count = LOOPS
            start_time = time.perf_counter()
            await trio.sleep(1)
            end_time = time.perf_counter()
            end_count = LOOPS
            loops = end_count - start_count
            duration = end_time - start_time
            print("{} loops/sec".format(loops / duration))
    finally:
        RUNNING = False

async def main():
    async with trio.open_nursery() as nursery:
        nursery.start_soon(reschedule_loop, 10)
        nursery.start_soon(report_loop)

trio.run(main)



import errno
import select

import os
import pytest

from trio._core.tests.tutil import gc_collect_harder
from ... import _core
from ...testing import (wait_all_tasks_blocked, check_one_way_stream)

posix = os.name == "posix"

pytestmark = pytest.mark.skipif(
    not posix, reason="pipes are only supported on posix"
)

if posix:
    from ..._subprocess.unix_pipes import (
        PipeSendStream, PipeReceiveStream, make_pipe
    )


async def test_send_pipe():
    r, w = os.pipe()
    send = PipeSendStream(w)
    assert send.fileno() == w
    await send.send_all(b"123")
    assert (os.read(r, 8)) == b"123"

    os.close(r)
    os.close(w)
    send._closed = True


async def test_receive_pipe():
    r, w = os.pipe()
    recv = PipeReceiveStream(r)
    assert (recv.fileno()) == r
    os.write(w, b"123")
    assert (await recv.receive_some(8)) == b"123"

    os.close(r)
    os.close(w)
    recv._closed = True


async def test_pipes_combined():
    write, read = await make_pipe()
    count = 2**20

    async def sender():
        big = bytearray(count)
        await write.send_all(big)

    async def reader():
        await wait_all_tasks_blocked()
        received = 0
        while received < count:
            received += len(await read.receive_some(4096))

        assert received == count

    async with _core.open_nursery() as n:
        n.start_soon(sender)
        n.start_soon(reader)

    await read.aclose()
    await write.aclose()


async def test_send_on_closed_pipe():
    write, read = await make_pipe()
    await write.aclose()

    with pytest.raises(_core.ClosedResourceError):
        await write.send_all(b"123")

    await read.aclose()


async def test_pipe_errors():
    with pytest.raises(TypeError):
        PipeReceiveStream(None)

    with pytest.raises(ValueError):
        await PipeReceiveStream(0).receive_some(0)


async def test_del():
    w, r = await make_pipe()
    f1, f2 = w.fileno(), r.fileno()
    del w, r
    gc_collect_harder()

    with pytest.raises(OSError) as excinfo:
        os.close(f1)
    assert excinfo.value.errno == errno.EBADF

    with pytest.raises(OSError) as excinfo:
        os.close(f2)
    assert excinfo.value.errno == errno.EBADF


async def test_async_with():
    w, r = await make_pipe()
    async with w, r:
        pass

    assert w._closed
    assert r._closed

    with pytest.raises(OSError) as excinfo:
        os.close(w.fileno())
    assert excinfo.value.errno == errno.EBADF

    with pytest.raises(OSError) as excinfo:
        os.close(r.fileno())
    assert excinfo.value.errno == errno.EBADF


async def make_clogged_pipe():
    s, r = await make_pipe()
    try:
        while True:
            # We want to totally fill up the pipe buffer.
            # This requires working around a weird feature that POSIX pipes
            # have.
            # If you do a write of <= PIPE_BUF bytes, then it's guaranteed
            # to either complete entirely, or not at all. So if we tried to
            # write PIPE_BUF bytes, and the buffer's free space is only
            # PIPE_BUF/2, then the write will raise BlockingIOError... even
            # though a smaller write could still succeed! To avoid this,
            # make sure to write >PIPE_BUF bytes each time, which disables
            # the special behavior.
            # For details, search for PIPE_BUF here:
            #   http://pubs.opengroup.org/onlinepubs/9699919799/functions/write.html

            # for the getattr:
            # https://bitbucket.org/pypy/pypy/issues/2876/selectpipe_buf-is-missing-on-pypy3
            buf_size = getattr(select, "PIPE_BUF", 8192)
            os.write(s.fileno(), b"x" * buf_size * 2)
    except BlockingIOError:
        pass
    return s, r


async def test_pipe_fully():
    await check_one_way_stream(make_pipe, make_clogged_pipe)



import socket

# Linux:
#   low values get rounded up to ~2-4 KB, so that's predictable
#   with low values, can queue up 6 one-byte sends (!)
#   with default values, can queue up 278 one-byte sends
#
# Windows:
#   if SNDBUF = 0 freezes, so that's useless
#   by default, buffers 655121
#   with both set to 1, buffers 525347
#   except sometimes it's less intermittently (?!?)
#
# macOS:
#   if bufsize = 1, can queue up 1 one-byte send
#   with default bufsize, can queue up 8192 one-byte sends
#   and bufsize = 0 is invalid (setsockopt errors out)

for bufsize in [1, None, 0]:
    a, b = socket.socketpair()
    a.setblocking(False)
    b.setblocking(False)

    a.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
    if bufsize is not None:
        a.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, bufsize)
        b.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, bufsize)

    try:
        for i in range(10000000):
            a.send(b"\x00")
    except BlockingIOError:
        pass

    print("setsockopt bufsize {}: {}".format(bufsize, i))
    a.close()
    b.close()



import os
import threading
import time
import tempfile

def check_reopen(r1, w):
    try:
        print("Reopening read end")
        r2 = os.open("/proc/self/fd/{}".format(r1), os.O_RDONLY)

        print("r1 is {}, r2 is {}".format(r1, r2))

        print("checking they both can receive from w...")

        os.write(w, b"a")
        assert os.read(r1, 1) == b"a"

        os.write(w, b"b")
        assert os.read(r2, 1) == b"b"

        print("...ok")

        print("setting r2 to non-blocking")
        os.set_blocking(r2, False)

        print("os.get_blocking(r1) ==", os.get_blocking(r1))
        print("os.get_blocking(r2) ==", os.get_blocking(r2))

        # Check r2 is really truly non-blocking
        try:
            os.read(r2, 1)
        except BlockingIOError:
            print("r2 definitely seems to be in non-blocking mode")

        # Check that r1 is really truly still in blocking mode
        def sleep_then_write():
            time.sleep(1)
            os.write(w, b"c")
        threading.Thread(target=sleep_then_write, daemon=True).start()
        assert os.read(r1, 1) == b"c"
        print("r1 definitely seems to be in blocking mode")
    except Exception as exc:
        print("ERROR: {!r}".format(exc))


print("-- testing anonymous pipe --")
check_reopen(*os.pipe())

print("-- testing FIFO --")
with tempfile.TemporaryDirectory() as tmpdir:
    fifo = tmpdir + "/" + "myfifo"
    os.mkfifo(fifo)
    # "A process can open a FIFO in nonblocking mode.  In this case, opening
    # for read-only will succeed even if no-one has opened on the write side
    # yet and opening for write-only will fail with ENXIO (no such device or
    # address) unless the other end has already been opened." -- Linux fifo(7)
    r = os.open(fifo, os.O_RDONLY | os.O_NONBLOCK)
    assert not os.get_blocking(r)
    os.set_blocking(r, True)
    assert os.get_blocking(r)
    w = os.open(fifo, os.O_WRONLY)
    check_reopen(r, w)

print("-- testing socketpair --")
import socket
rs, ws = socket.socketpair()
check_reopen(rs.fileno(), ws.fileno())




# what does SO_REUSEADDR do, exactly?

# Theory:
#
# - listen1 is bound to port P
# - listen1.accept() returns a connected socket server1, which is also bound
#   to port P
# - listen1 is closed
# - we attempt to bind listen2 to port P
# - this fails because server1 is still open, or still in TIME_WAIT, and you
#   can't use bind() to bind to a port that still has sockets on it, unless
#   both those sockets and the socket being bound have SO_REUSEADDR
#
# The standard way to avoid this is to set SO_REUSEADDR on all listening
# sockets before binding them. And this works, but for somewhat more
# complicated reasons than are often appreciated.
#
# In our scenario above it doesn't really matter for listen1 (assuming the
# port is initially unused).
#
# What is important is that it's set on *server1*. Setting it on listen1
# before calling bind() automatically accomplishes this, because SO_REUSEADDR
# is inherited by accept()ed sockets. But it also works to set it on listen1
# any time before calling accept(), or to set it on server1 directly.
#
# Also, it must be set on listen2 before calling bind(), or it will conflict
# with the lingering server1 socket.

import socket
import errno

import attr

@attr.s(repr=False)
class Options:
    listen1_early = attr.ib(default=None)
    listen1_middle = attr.ib(default=None)
    listen1_late = attr.ib(default=None)
    server = attr.ib(default=None)
    listen2 = attr.ib(default=None)

    def set(self, which, sock):
        value = getattr(self, which)
        if value is not None:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, value)

    def describe(self):
        info = []
        for f in attr.fields(self.__class__):
            value = getattr(self, f.name)
            if value is not None:
                info.append("{}={}".format(f.name, value))
        return "Set/unset: {}".format(", ".join(info))

def time_wait(options):
    print(options.describe())

    # Find a pristine port (one we can definitely bind to without
    # SO_REUSEADDR)
    listen0 = socket.socket()
    listen0.bind(("127.0.0.1", 0))
    sockaddr = listen0.getsockname()
    #print("  ", sockaddr)
    listen0.close()

    listen1 = socket.socket()
    options.set("listen1_early", listen1)
    listen1.bind(sockaddr)
    listen1.listen(1)

    options.set("listen1_middle", listen1)

    client = socket.socket()
    client.connect(sockaddr)

    options.set("listen1_late", listen1)

    server, _ = listen1.accept()

    options.set("server", server)

    # Server initiated close to trigger TIME_WAIT status
    server.close()
    assert client.recv(10) == b""
    client.close()

    listen1.close()

    listen2 = socket.socket()
    options.set("listen2", listen2)
    try:
        listen2.bind(sockaddr)
    except OSError as exc:
        if exc.errno == errno.EADDRINUSE:
            print("  -> EADDRINUSE")
        else:
            raise
    else:
        print("  -> ok")

time_wait(Options())
time_wait(Options(listen1_early=True, server=True, listen2=True))
time_wait(Options(listen1_early=True))
time_wait(Options(server=True))
time_wait(Options(listen2=True))
time_wait(Options(listen1_early=True, listen2=True))
time_wait(Options(server=True, listen2=True))
time_wait(Options(listen1_middle=True, listen2=True))
time_wait(Options(listen1_late=True, listen2=True))
time_wait(Options(listen1_middle=True, server=False, listen2=True))



# This is a public namespace, so we don't want to expose any non-underscored
# attributes that aren't actually part of our public API. But it's very
# annoying to carefully always use underscored names for module-level
# temporaries, imports, etc. when implementing the module. So we put the
# implementation in an underscored module, and then re-export the public parts
# here.
# We still have some underscore names though but only a few.

from . import _socket
import sys as _sys

# import the overwrites
from ._socket import (
    fromfd, from_stdlib_socket, getprotobyname, socketpair, getnameinfo,
    socket, getaddrinfo, set_custom_hostname_resolver,
    set_custom_socket_factory, SocketType
)

# not always available so expose only if
try:
    from ._socket import fromshare
except ImportError:
    pass

# expose these functions to trio.socket
from socket import (
    gaierror,
    herror,
    gethostname,
    ntohs,
    htonl,
    htons,
    inet_aton,
    inet_ntoa,
    inet_pton,
    inet_ntop,
)

# not always available so expose only if
try:
    from socket import (
        sethostname, if_nameindex, if_nametoindex, if_indextoname
    )
except ImportError:
    pass

# expose all uppercase names from standardlib socket to trio.socket
import socket as _stdlib_socket

globals().update(
    {
        _name: getattr(_stdlib_socket, _name)
        for _name in _stdlib_socket.__dict__ if _name.isupper()
    }
)

if _sys.platform == 'win32':
    # See https://github.com/python-trio/trio/issues/39
    # Do not import for windows platform
    # (you can still get it from stdlib socket, of course, if you want it)
    del SO_REUSEADDR

# get names used by trio that we define on our own
from ._socket import IPPROTO_IPV6

# Not defined in all python versions and platforms but sometimes needed
try:
    TCP_NOTSENT_LOWAT
except NameError:
    # Hopefully will show up in 3.7:
    #   https://github.com/python/cpython/pull/477
    if _sys.platform == "darwin":
        TCP_NOTSENT_LOWAT = 0x201
    elif _sys.platform == "linux":
        TCP_NOTSENT_LOWAT = 25



import attr
import functools
import os
import outcome
from typing import Any

from .. import _core
from .._sync import Event
from .._threads import run_sync_in_worker_thread


@attr.s
class WaitpidState:
    pid = attr.ib()
    event = attr.ib(default=attr.Factory(Event))
    outcome = attr.ib(default=None)


# https://github.com/python-trio/trio/issues/618
class StubLimiter:
    def release_on_behalf_of(self, x):
        pass

    async def acquire_on_behalf_of(self, x):
        pass


waitpid_limiter = StubLimiter()


# adapted from
# https://github.com/python-trio/trio/issues/4#issuecomment-398967572
async def _task(state: WaitpidState) -> None:
    """The waitpid thread runner task. This must be spawned as a system
    task."""
    partial = functools.partial(
        os.waitpid,  # function
        state.pid,  # pid
        0  # no options
    )

    tresult = await run_sync_in_worker_thread(
        outcome.capture, partial, cancellable=True, limiter=waitpid_limiter
    )
    state.outcome = tresult
    state.event.set()


async def waitpid(pid: int) -> Any:
    """Waits for a child process with the specified PID to finish running."""
    waiter = WaitpidState(pid=pid)
    _core.spawn_system_task(_task, waiter)

    await waiter.event.wait()
    return waiter.outcome.unwrap()



import outcome

from .. import _deprecate

__all__ = ["Result", "Value", "Error"]


class Result(outcome.Outcome):
    @classmethod
    @_deprecate.deprecated(
        version="0.5.0", issue=494, instead="outcome.capture"
    )
    def capture(cls, sync_fn, *args):
        return outcome.capture(sync_fn, *args)

    @classmethod
    @_deprecate.deprecated(
        version="0.5.0", issue=494, instead="outcome.acapture"
    )
    async def acapture(cls, async_fn, *args):
        return await outcome.acapture(async_fn, *args)


# alias these types so they don't mysteriously disappear
Value = outcome.Value
Error = outcome.Error

# ensure that isinstance(Value(), Result)/issubclass(Value, Result) and etc
# don't break
Result.register(Value)
Result.register(Error)
